/********************************************************************
* Multithreaded SDI Applications, version 1.5 (October 5, 2002)
* Copyright (C) 2002-2003 Michal Mecinski.
*
* You may freely use and modify this code, but don't remove
* this copyright note.
*
* THERE IS NO WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, FOR
* THIS CODE. THE AUTHOR DOES NOT TAKE THE RESPONSIBILITY
* FOR ANY DAMAGE RESULTING FROM THE USE OF IT.
*
* E-mail: mimec@mimec.org
* WWW: http://www.mimec.org
********************************************************************/

#include "stdafx.h"
#include <afxpriv.h>
#include <dde.h>
#include "MSDI.h"

#include "MSDIDocManager.h"
#include "MSDIThread.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


/////////////////////////////////////////////////////////////////////////////
// CMSDIApp

BEGIN_MESSAGE_MAP(CMSDIApp, CWinApp)
	//{{AFX_MSG_MAP(CMSDIApp)
	ON_COMMAND(ID_APP_ABOUT, OnAppAbout)
	ON_COMMAND(ID_APP_EXIT, OnAppExit)
	//}}AFX_MSG_MAP
	// Standard file based document commands
	ON_COMMAND(ID_FILE_NEW, CWinApp::OnFileNew)
	ON_COMMAND(ID_FILE_OPEN, CWinApp::OnFileOpen)
	// Standard print setup command
	ON_COMMAND(ID_FILE_PRINT_SETUP, CWinApp::OnFilePrintSetup)
	// Handle thread messages
	ON_THREAD_MESSAGE(MSDIM_EXIT_THREAD, OnExitThread)
	ON_THREAD_MESSAGE(MSDIM_NEW_INSTANCE, OnNewInstance)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CMSDIApp construction

CMSDIApp::CMSDIApp()
{
	m_pFrameWnd = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// The one and only CMSDIApp object

CMSDIApp theApp;

/////////////////////////////////////////////////////////////////////////////
// CMSDIApp initialization

BOOL CMSDIApp::InitInstance()
{
	if (!ParseCommandLine())
		return FALSE;

	if (IsAlreadyRunning("Unique App Name"))
		return FALSE;

	SetRegistryKey(_T("MSDI Application"));

	LoadStdProfileSettings();

	m_pDocManager = new CMSDIDocManager;

	EnableShellOpen();

	if (!ProcessShellCommand())
		return FALSE;

	return TRUE;
}


/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
		// No message handlers
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
		// No message handlers
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

// App command to run the dialog
void CMSDIApp::OnAppAbout()
{
	CAboutDlg aboutDlg;
	aboutDlg.DoModal();
}

/////////////////////////////////////////////////////////////////////////////
// CMSDIApp message handlers

int CMSDIApp::Run() 
{
	return CWinThread::Run();
}

int CMSDIApp::ExitInstance() 
{
	if (m_pFrameWnd)
		m_pFrameWnd->DestroyWindow();

	if (m_pDocManager)
		((CMSDIDocManager*)m_pDocManager)->KillAllThreads();

	if (!(m_fCommand & CMD_UNREGSERVER))
		SaveStdProfileSettings();
	
	return 0;
}


void CMSDIApp::OnExitThread(WPARAM pThread, LPARAM)
{
	((CMSDIDocManager*)m_pDocManager)->OnExitThread((CMSDIThread*)pThread);
}

void CMSDIApp::OnNewInstance(WPARAM nFileOpen, LPARAM)
{
	if (nFileOpen)
	{
		char szBuf[MAX_PATH];

		if (GlobalGetAtomName((ATOM)nFileOpen, szBuf, sizeof(szBuf)))
			OpenDocumentFile(szBuf);

		GlobalDeleteAtom((ATOM)nFileOpen);
	}
	else
		OnFileNew();
}


BOOL CMSDIApp::ParseCommandLine()
{
	for (int i=1; i < __argc; i++)
	{
		LPCSTR lpszParam = __argv[i];
		
		if (lpszParam[0] == '-' || lpszParam[0] == '/')
		{
			lpszParam++;
			if (!strcmp(lpszParam, "dde"))
				m_fCommand |= CMD_DDE;
			else if (!stricmp(lpszParam, "nosplash"))
				m_fCommand |= CMD_NOSPLASH;
			else if (!stricmp(lpszParam, "regserver"))
				m_fCommand = CMD_REGSERVER;
			else if (!stricmp(lpszParam, "unregserver"))
				m_fCommand = CMD_UNREGSERVER;
		}
		else if (!(m_fCommand & CMD_OPEN))
		{
			m_strFileOpen = lpszParam;
			m_fCommand |= CMD_OPEN;
		}
	}

	if (m_fCommand & CMD_UNREGSERVER)
		Unregister();
	else
		Register();

	return !(m_fCommand & (CMD_REGSERVER | CMD_UNREGSERVER));
}

BOOL CMSDIApp::ProcessShellCommand()
{
	if (m_fCommand & CMD_DDE)
		return TRUE;

	if (m_fCommand & CMD_OPEN)
		OpenDocumentFile(m_strFileOpen);
	else
		OnFileNew();

	if (((CMSDIDocManager*)m_pDocManager)->IsThreadListEmpty())
		return FALSE;

	return TRUE;
}


BOOL CMSDIApp::IsAlreadyRunning(LPCSTR lpszName)
{
	ASSERT(lpszName!=NULL);

	LPCSTR lpszClassName = AfxRegisterWndClass(0, NULL, NULL, NULL);

	HWND hWnd = FindWindow(lpszClassName, lpszName);

	if (hWnd)
	{
		ATOM nFileOpen = NULL;

		if (m_fCommand & CMD_OPEN)
			nFileOpen = GlobalAddAtom(m_strFileOpen);

		DWORD dwThreadID = GetWindowThreadProcessId(hWnd, NULL);

		::PostThreadMessage(dwThreadID, MSDIM_NEW_INSTANCE, (WPARAM)nFileOpen, 0);

		return TRUE;
	}

	m_pFrameWnd = (CFrameWnd*)RUNTIME_CLASS(CFrameWnd)->CreateObject();

	m_pFrameWnd->Create(lpszClassName, lpszName);

	return FALSE;
}


BOOL CMSDIApp::Register()
{
	CString strShort;
	AfxGetModuleShortFileName(m_hInstance, strShort);

	CString strDescr;
	strDescr.LoadString(IDR_MAINFRAME);

	CString strExt, strID, strName;
	AfxExtractSubString(strExt, strDescr, CDocTemplate::filterExt, '\n');
	AfxExtractSubString(strID, strDescr, CDocTemplate::regFileTypeId, '\n');
	AfxExtractSubString(strName, strDescr, CDocTemplate::regFileTypeName, '\n');

	SetRegKey(strExt, strID);
	SetRegKey(strID, strName);
	SetRegKey(strID + "\\DefaultIcon", strShort + ",1");

	SetRegKey(strID + "\\shell\\open\\command", strShort + " /dde");
	SetRegKey(strID + "\\shell\\open\\ddeexec", "[open(\"%1\")]");

	return TRUE;
}

BOOL CMSDIApp::Unregister()
{
	CWinApp::Unregister();

	CString strDescr;
	strDescr.LoadString(IDR_MAINFRAME);

	CString strExt, strID;
	AfxExtractSubString(strExt, strDescr, CDocTemplate::filterExt, '\n');
	AfxExtractSubString(strID, strDescr, CDocTemplate::regFileTypeId, '\n');

	DelRegTree(HKEY_CLASSES_ROOT, strExt);
	DelRegTree(HKEY_CLASSES_ROOT, strID);

	return TRUE;
}

void CMSDIApp::SetRegKey(LPCSTR lpszKey, LPCSTR lpszValue)
{
	RegSetValue(HKEY_CLASSES_ROOT, lpszKey, REG_SZ,
		  lpszValue, strlen(lpszValue));
}


BOOL CMSDIApp::OnCmdMsg(UINT nID, int nCode, void* pExtra, AFX_CMDHANDLERINFO* pHandlerInfo) 
{
	CMSDIThread* pThread = (CMSDIThread*)AfxGetThread();
	
	if (pThread->IsKindOf(RUNTIME_CLASS(CMSDIThread)))
	{
		if (pThread->OnCmdMsg(nID, nCode, pExtra, pHandlerInfo))
			return TRUE;
	}
	return CWinApp::OnCmdMsg(nID, nCode, pExtra, pHandlerInfo);
}


void CMSDIApp::OnAppExit() 
{
	((CMSDIDocManager*)m_pDocManager)->CloseAllFrames();
}


